import platform

APOLLO_PLATFORMS = {
    (0, 1): 'cynthion.gateware.platform:CynthionPlatformRev0D1',
    (0, 2): 'cynthion.gateware.platform:CynthionPlatformRev0D2',
    (0, 3): 'cynthion.gateware.platform:CynthionPlatformRev0D3',
    (0, 4): 'cynthion.gateware.platform:CynthionPlatformRev0D4',
    (0, 5): 'cynthion.gateware.platform:CynthionPlatformRev0D5',
    (0, 6): 'cynthion.gateware.platform:CynthionPlatformRev0D6',
    (0, 7): 'cynthion.gateware.platform:CynthionPlatformRev0D7',
    (1, 0): 'cynthion.gateware.platform:CynthionPlatformRev1D0',
    (1, 1): 'cynthion.gateware.platform:CynthionPlatformRev1D1',
    (1, 2): 'cynthion.gateware.platform:CynthionPlatformRev1D2',
    (1, 3): 'cynthion.gateware.platform:CynthionPlatformRev1D3',
    (1, 4): 'cynthion.gateware.platform:CynthionPlatformRev1D4',
}
