#!/usr/bin/env python3
#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

REGISTER_ID = 1
REGISTER_LEDS = 2

REGISTER_TARGET_ADDR = 7
REGISTER_TARGET_VALUE = 8
REGISTER_TARGET_RXCMD = 9

REGISTER_AUX_ADDR = 10
REGISTER_AUX_VALUE = 11
REGISTER_AUX_RXCMD = 12

REGISTER_CONTROL_ADDR = 13
REGISTER_CONTROL_VALUE = 14
REGISTER_CONTROL_RXCMD = 15

REGISTER_RAM_REG_ADDR = 20
REGISTER_RAM_VALUE = 21
