#
# This file is part of Cynthion.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from .cynthion_r0_1 import CynthionPlatformRev0D1
from .cynthion_r0_2 import CynthionPlatformRev0D2
from .cynthion_r0_3 import CynthionPlatformRev0D3
from .cynthion_r0_4 import CynthionPlatformRev0D4
from .cynthion_r0_5 import CynthionPlatformRev0D5
from .cynthion_r0_6 import CynthionPlatformRev0D6
from .cynthion_r0_7 import CynthionPlatformRev0D7
from .cynthion_r1_0 import CynthionPlatformRev1D0
from .cynthion_r1_1 import CynthionPlatformRev1D1
from .cynthion_r1_2 import CynthionPlatformRev1D2
from .cynthion_r1_3 import CynthionPlatformRev1D3
from .cynthion_r1_4 import CynthionPlatformRev1D4
